from pyhafas.profile import ProfileInterface
from pyhafas.profile.interfaces.helper.format_products_filter import \
    FormatProductsFilterHelperInterface
from pyhafas.types.exceptions import ProductNotAvailableError


class BaseFormatProductsFilterHelper(FormatProductsFilterHelperInterface):
    def format_products_filter(
            self: ProfileInterface,
            requested_products: dict) -> dict:
        """
        Create the products filter given to HaFAS

        :param requested_products: Mapping of Products to whether it's enabled or disabled
        :return:
        """
        products = self.default_products
        for requested_product in requested_products:
            if requested_products[requested_product]:
                try:
                    products.index(requested_product)
                except ValueError:
                    products.append(requested_product)

            elif not requested_products[requested_product]:
                try:
                    products.pop(products.index(requested_product))
                except ValueError:
                    pass
        bitmask_sum = 0
        for product in products:
            try:
                for product_bitmask in self.available_products[product]:
                    bitmask_sum += product_bitmask
            except KeyError:
                raise ProductNotAvailableError(
                    'The product "{}" is not available in chosen profile.'.format(product))
        return {
            'type': 'PROD',
            'mode': 'INC',
            'value': str(bitmask_sum)
        }



import abc

from pyhafas.types.fptf import Station


class ParseLidHelperInterface(abc.ABC):
    @abc.abstractmethod
    def parse_lid(self, lid: str) -> dict:
        pass

    @abc.abstractmethod
    def parse_lid_to_station(
            self,
            lid: str,
            name: str = "",
            latitude: int = 0,
            longitude: int = 0) -> Station:
        pass



from typing import Dict, List

from pyhafas.profile.base.helper.format_products_filter import \
    BaseFormatProductsFilterHelper
from pyhafas.profile.base.helper.parse_date_time import BaseParseDateTimeHelper
from pyhafas.profile.base.helper.parse_leg import BaseParseLegHelper
from pyhafas.profile.base.helper.parse_lid import BaseParseLidHelper
from pyhafas.profile.base.helper.request import BaseRequestHelper
from pyhafas.profile.base.requests.journey import BaseJourneyRequest
from pyhafas.profile.base.requests.journeys import BaseJourneysRequest
from pyhafas.profile.base.requests.location import BaseLocationRequest
from pyhafas.profile.base.requests.station_board import BaseStationBoardRequest
from pyhafas.profile.base.requests.trip import BaseTripRequest
from pyhafas.profile.interfaces import ProfileInterface


class BaseProfile(
        BaseRequestHelper,
        BaseFormatProductsFilterHelper,
        BaseParseLidHelper,
        BaseParseDateTimeHelper,
        BaseParseLegHelper,
        BaseLocationRequest,
        BaseJourneyRequest,
        BaseJourneysRequest,
        BaseStationBoardRequest,
        BaseTripRequest,
        ProfileInterface):
    baseUrl: str = ""
    defaultUserAgent: str = 'pyhafas'

    addMicMac: bool = False
    addChecksum: bool = False
    salt: str = ""

    locale: str = 'de-DE'
    timezone: str = 'Europe/Berlin'

    requestBody: dict = {}

    available_products: Dict[str, List[int]] = {}
    default_products: List[str] = []

    def __init__(self, ua=defaultUserAgent):
        self.userAgent = ua



import abc
import datetime
from typing import List

from pyhafas.types.fptf import Leg


class ParseLegHelperInterface(abc.ABC):
    @abc.abstractmethod
    def parse_leg(
            self,
            journey: dict,
            common: dict,
            departure: dict,
            arrival: dict,
            date: datetime.date,
            jny_type: str = "JNY",
            gis=None) -> Leg:
        pass

    @abc.abstractmethod
    def parse_legs(
            self,
            jny: dict,
            common: dict,
            date: datetime.date) -> List[Leg]:
        pass



from abc import ABC
from typing import Dict, List

from pyhafas.profile.interfaces.helper.format_products_filter import \
    FormatProductsFilterHelperInterface
from pyhafas.profile.interfaces.helper.parse_date_time import \
    ParseDateTimeHelperInterface
from pyhafas.profile.interfaces.helper.parse_leg import ParseLegHelperInterface
from pyhafas.profile.interfaces.helper.parse_lid import ParseLidHelperInterface
from pyhafas.profile.interfaces.helper.request import RequestHelperInterface
from pyhafas.profile.interfaces.requests.journey import JourneyRequestInterface
from pyhafas.profile.interfaces.requests.journeys import \
    JourneysRequestInterface
from pyhafas.profile.interfaces.requests.location import \
    LocationRequestInterface
from pyhafas.profile.interfaces.requests.station_board import \
    StationBoardRequestInterface
from pyhafas.profile.interfaces.requests.trip import TripRequestInterface


class ProfileInterface(
        RequestHelperInterface,
        FormatProductsFilterHelperInterface,
        ParseLidHelperInterface,
        ParseDateTimeHelperInterface,
        ParseLegHelperInterface,
        LocationRequestInterface,
        JourneyRequestInterface,
        JourneysRequestInterface,
        StationBoardRequestInterface,
        TripRequestInterface,
        ABC):
    baseUrl: str
    defaultUserAgent: str
    userAgent: str

    addMicMac: bool
    addChecksum: bool
    salt: str

    locale: str
    timezone: str

    requestBody: dict

    available_products: Dict[str, List[int]]
    default_products: List[str]



import abc
import datetime
from typing import Dict, List

from pyhafas.types.fptf import Journey, Station
from pyhafas.types.hafas_response import HafasResponse


class JourneysRequestInterface(abc.ABC):
    @abc.abstractmethod
    def format_journeys_request(
            self,
            origin: Station,
            destination: Station,
            via: List[Station],
            date: datetime.datetime,
            min_change_time: int,
            max_changes: int,
            products: Dict[str, bool]
    ) -> dict:
        pass

    @abc.abstractmethod
    def parse_journeys_request(self, data: HafasResponse) -> List[Journey]:
        pass



import abc
from typing import List

from pyhafas.types.fptf import Station
from pyhafas.types.hafas_response import HafasResponse


class LocationRequestInterface(abc.ABC):
    @abc.abstractmethod
    def format_location_request(self, term: str):
        pass

    @abc.abstractmethod
    def parse_location_request(self, data: HafasResponse) -> List[Station]:
        pass



import abc
import datetime
from typing import Dict, List

from pyhafas.types.fptf import Leg, Station, StationBoardRequestType
from pyhafas.types.hafas_response import HafasResponse


class StationBoardRequestInterface(abc.ABC):
    def format_station_board_request(
            self,
            station: Station,
            request_type: StationBoardRequestType,
            date: datetime.datetime,
            max_journeys: int,
            duration: int,
            products: Dict[str, bool]
    ) -> dict:
        pass

    def parse_station_board_request(self, data: HafasResponse) -> List[Leg]:
        pass



import abc
from typing import Tuple

from pyhafas.types.hafas_response import HafasResponse


class RequestHelperInterface(abc.ABC):
    @abc.abstractmethod
    def calculate_checksum(self, data: str) -> str:
        pass

    @abc.abstractmethod
    def calculate_mic_mac(self, data: str) -> Tuple[str, str]:
        pass

    @abc.abstractmethod
    def url_formatter(self, data: str) -> str:
        pass

    @abc.abstractmethod
    def request(self, body) -> HafasResponse:
        pass



import abc
import datetime


class ParseDateTimeHelperInterface(abc.ABC):
    @abc.abstractmethod
    def parse_datetime(
            self,
            time_string: str,
            date: datetime.date) -> datetime.datetime:
        pass

    @abc.abstractmethod
    def parse_timedelta(self, time_string: str) -> datetime.timedelta:
        pass

    @abc.abstractmethod
    def parse_date(self, date_string: str) -> datetime.date:
        pass



import datetime
from enum import Enum
from typing import List, Optional, Union


class StationBoardRequestType(Enum):
    DEPARTURE = 'DEP'
    ARRIVAL = 'ARR'

    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)


class Mode(Enum):
    TRAIN = 'train'
    BUS = 'bus'
    WATERCRAFT = 'watercraft'
    TAXI = 'taxi'
    GONDOLA = 'gondola'
    AIRCRAFT = 'aircraft'
    CAR = 'car'
    BICYCLE = 'bicycle'
    WALKING = 'walking'

    def __repr__(self):
        return '<%s.%s>' % (self.__class__.__name__, self.name)


class Station:
    def __init__(self, id: Union[str, int], **kwargs):
        self.id = id
        self.name: Optional[str] = kwargs.get('name')
        self.latitude: Optional[float] = kwargs.get('latitude')
        self.longitude: Optional[float] = kwargs.get('longitude')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class Journey:
    def __init__(self, id: Union[str, int], **kwargs):
        self.id = id
        self.date: Optional[datetime.date] = kwargs.get('date')
        self.duration: Optional[datetime.timedelta] = kwargs.get('duration')
        self.legs: Optional[List[Leg]] = kwargs.get('legs')

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class Leg:
    def __init__(self, id: Union[str, int], **kwargs):
        self.id = id
        self.name: Optional[str] = kwargs.get('name')
        self.origin: Station = kwargs['origin']
        self.mode: Mode = kwargs.get('mode', Mode.TRAIN)
        self.cancelled: bool = kwargs.get('cancelled', False)
        self.destination: Station = kwargs['destination']
        self.departure: datetime.datetime = kwargs['departure']
        self.departureDelay: Optional[datetime.timedelta] = kwargs.get(
            'departureDelay', None)
        self.departurePlatform: Optional[str] = kwargs.get(
            'departurePlatform', None)
        self.arrival: datetime.datetime = kwargs['arrival']
        self.arrivalDelay: Optional[datetime.timedelta] = kwargs.get(
            'arrivalDelay', None)
        self.arrivalPlatform: Optional[str] = kwargs.get(
            'arrivalPlatform', None)
        self.stopovers: Optional[List[Stopover]
                                 ] = kwargs.get('stopovers', None)
        self.distance: Optional[int] = kwargs.get('distance', None)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)


class Stopover:
    def __init__(self, **kwargs):
        self.stop: Station = kwargs.get('stop')
        self.cancelled: bool = kwargs.get('cancelled', False)
        self.arrival: Optional[datetime.datetime] = kwargs.get('arrival', None)
        self.arrivalDelay: Optional[datetime.timedelta] = kwargs.get(
            'arrivalDelay', None)
        self.arrivalPlatform: Optional[str] = kwargs.get(
            'arrivalPlatform', None)
        self.departure: Optional[datetime.datetime] = kwargs.get(
            'departure', None)
        self.departureDelay: Optional[datetime.timedelta] = kwargs.get(
            'departureDelay', None)
        self.departurePlatform: Optional[str] = kwargs.get(
            'departurePlatform', None)

    def __repr__(self):
        return "%s(%r)" % (self.__class__, self.__dict__)



import json
from hashlib import md5
from typing import Tuple

import requests

from pyhafas.profile import ProfileInterface
from pyhafas.profile.base.mappings.error_codes import BaseErrorCodesMapping
from pyhafas.profile.interfaces.helper.request import RequestHelperInterface
from pyhafas.types.hafas_response import HafasResponse


class BaseRequestHelper(RequestHelperInterface):
    def calculate_checksum(self: ProfileInterface, data: str) -> str:
        return md5((data + self.salt).encode('utf-8')).hexdigest()

    def calculate_mic_mac(
            self: ProfileInterface,
            data: str) -> Tuple[str, str]:
        mic = md5(data.encode('utf-8')).hexdigest()
        mac = self.calculate_checksum(mic)
        return mic, mac

    def url_formatter(self: ProfileInterface, data: str) -> str:
        url = self.baseUrl

        if self.addChecksum or self.addMicMac:
            parameters = []
            if self.addChecksum:
                parameters.append(
                    'checksum={}'.format(
                        self.calculate_checksum(data)))
            if self.addMicMac:
                parameters.append(
                    'mic={}&mac={}'.format(
                        *self.calculate_mic_mac(data)))
            url += '?{}'.format('&'.join(parameters))

        return url

    def request(self: ProfileInterface, body) -> HafasResponse:
        data = {
            'svcReqL': [body]
        }
        data.update(self.requestBody)
        data = json.dumps(data)

        res = requests.post(
            self.url_formatter(data),
            data=data,
            headers={
                'User-Agent': self.userAgent,
                'Content-Type': 'application/json'})
        return HafasResponse(res, BaseErrorCodesMapping)



import abc

from pyhafas.types.fptf import Leg
from pyhafas.types.hafas_response import HafasResponse


class TripRequestInterface(abc.ABC):
    @abc.abstractmethod
    def format_trip_request(self, trip_id: str) -> dict:
        pass

    @abc.abstractmethod
    def parse_trip_request(self, data: HafasResponse) -> Leg:
        pass



import datetime
from typing import Dict, List

from .profile import BaseProfile, DBProfile
from .types.fptf import Journey, Leg, Station, StationBoardRequestType


class HafasClient:
    def __init__(
            self,
            profile: BaseProfile = DBProfile(),
            ua: str = "pyhafas",
            debug: bool = False):

        self.profile = profile
        self.useragent = ua
        self.debug = debug

    def departures(
            self,
            station,
            date: datetime.datetime,
            max_journeys: int = -1,
            duration: int = -1,
            products: Dict[str, bool] = {}) -> List[Leg]:
        if not isinstance(station, Station):
            station = Station(station)

        body = self.profile.format_station_board_request(
            station,
            StationBoardRequestType.DEPARTURE,
            date,
            max_journeys,
            duration,
            products
        )
        res = self.profile.request(body)

        return self.profile.parse_station_board_request(res)

    def arrivals(
            self,
            station,
            date: datetime.datetime,
            max_journeys: int = -1,
            duration: int = -1,
            products: Dict[str, bool] = {}) -> List[Leg]:
        if not isinstance(station, Station):
            station = Station(station)

        body = self.profile.format_station_board_request(
            station,
            StationBoardRequestType.ARRIVAL,
            date,
            max_journeys,
            duration,
            products
        )
        res = self.profile.request(body)

        return self.profile.parse_station_board_request(res)

    def journeys(
            self,
            origin,
            destination,
            date: datetime.datetime,
            via: List = [],
            min_change_time: int = 0,
            max_changes: int = -1,
            products: Dict[str, bool] = {}
    ) -> List[Journey]:
        if not isinstance(origin, Station):
            origin = Station(origin)
        if not isinstance(destination, Station):
            destination = Station(destination)
        for via_station in via:
            if not isinstance(via_station, Station):
                via[via.index(via_station)] = Station(via_station)

        body = self.profile.format_journeys_request(
            origin,
            destination,
            via,
            date,
            min_change_time,
            max_changes,
            products
        )
        res = self.profile.request(body)

        return self.profile.parse_journeys_request(res)

    def journey(self, journey) -> Journey:
        if not isinstance(journey, Journey):
            journey = Journey(journey)

        body = self.profile.format_journey_request(journey)
        res = self.profile.request(body)

        return self.profile.parse_journey_request(res)

    def locations(self, term: str) -> List[Station]:
        body = self.profile.format_location_request(term)
        res = self.profile.request(body)

        return self.profile.parse_location_request(res)

    def trip(self, id: str):
        body = self.profile.format_trip_request(id)
        res = self.profile.request(body)

        return self.profile.parse_trip_request(res)

    def stop(self, stop):
        pass

    def nearby(self, location):
        pass

    def radar(self, north, west, south, east):
        pass



from pyhafas.profile import ProfileInterface
from pyhafas.profile.interfaces.helper.parse_lid import ParseLidHelperInterface
from pyhafas.types.fptf import Station


class BaseParseLidHelper(ParseLidHelperInterface):
    def parse_lid(self: ProfileInterface, lid: str) -> dict:
        """
        Converts the LID given by HaFAS

        Splits the LID (e.g. A=1@O=Siegburg/Bonn) in multiple elements (e.g. A=1 and O=Siegburg/Bonn).
        These are converted into a dict where the part before the equal sign is the key and the part after the value.

        :param lid: Location identifier (given by HaFAS)
        :return: Dict of the elements of the dict
        """
        parsedLid = {}
        for lidElementGroup in lid.split("@"):
            if lidElementGroup:
                parsedLid[lidElementGroup.split(
                    "=")[0]] = lidElementGroup.split("=")[1]
        return parsedLid

    def parse_lid_to_station(
            self: ProfileInterface,
            lid: str,
            name: str = "",
            latitude: int = 0,
            longitude: int = 0) -> Station:
        """
        Parses the LID given by HaFAS to a station object

        :param lid: Location identifier (given by HaFAS)
        :param name: Station name (optional, if not given, LID is used)
        :param latitude: Latitude of the station (optional, if not given, LID is used)
        :param longitude: Longitude of the station (optional, if not given, LID is used)
        :return: Station object
        """
        parsedLid = self.parse_lid(lid)
        if latitude == 0 and longitude == 0 and parsedLid['X'] and parsedLid['Y']:
            latitude = int(int(parsedLid['Y']) / 1000000)
            longitude = int(int(parsedLid['X']) / 1000000)

        return Station(
            id=parsedLid['L'],
            name=name or parsedLid['O'],
            latitude=latitude,
            longitude=longitude
        )



import datetime

from pyhafas.profile import ProfileInterface
from pyhafas.profile.interfaces.helper.parse_date_time import \
    ParseDateTimeHelperInterface


class BaseParseDateTimeHelper(ParseDateTimeHelperInterface):
    def parse_datetime(
            self: ProfileInterface,
            time_string: str,
            date: datetime.date) -> datetime.datetime:
        try:
            hour = int(time_string[-6:-4])
            minute = int(time_string[-4:-2])
            second = int(time_string[-2:])
        except ValueError:
            raise ValueError(
                'Time string "{}" has wrong format'.format(time_string))

        dateOffset = int(time_string[:2]) if len(time_string) > 6 else 0
        return datetime.datetime(
            date.year,
            date.month,
            date.day,
            hour,
            minute,
            second) + datetime.timedelta(days=dateOffset)

    def parse_timedelta(
            self: ProfileInterface,
            time_string: str) -> datetime.timedelta:
        try:
            hours = int(time_string[:2])
            minutes = int(time_string[2:-2])
            seconds = int(time_string[-2:])
        except ValueError:
            raise ValueError(
                'Time string "{}" has wrong format'.format(time_string))

        return datetime.timedelta(
            hours=hours,
            minutes=minutes,
            seconds=seconds)

    def parse_date(self: ProfileInterface, date_string: str) -> datetime.date:
        dt = datetime.datetime.strptime(date_string, '%Y%m%d')
        return dt.date()



import datetime
from typing import Dict, List

from pyhafas.profile import ProfileInterface
from pyhafas.profile.interfaces.requests.station_board import \
    StationBoardRequestInterface
from pyhafas.types.fptf import Leg, Station, StationBoardRequestType
from pyhafas.types.hafas_response import HafasResponse


class BaseStationBoardRequest(StationBoardRequestInterface):
    def format_station_board_request(
            self: ProfileInterface,
            station: Station,
            request_type: StationBoardRequestType,
            date: datetime.datetime,
            max_journeys: int,
            duration: int,
            products: Dict[str, bool]
    ) -> dict:
        """
        Creates the HaFAS request for Station Board (departure/arrival)

        :param station: Station to get departures/arrivals for
        :param request_type: ARRIVAL or DEPARTURE
        :param date: Date and time to get departures/arrival for
        :param max_journeys: Maximum number of trips that can be returned
        :param products: Allowed products (e.g. ICE,IC)
        :param duration: Time in which trips are searched
        :return: Request for HaFAS
        """
        # TODO: More options
        return {
            'req': {
                'type': request_type.value,
                'stbLoc': {
                    'lid': 'A=1@L={}@'.format(station.id)
                },
                'maxJny': max_journeys,
                'date': date.strftime("%Y%m%d"),
                'time': date.strftime("%H%M%S"),
                'dur': duration,
                'jnyFltrL': [
                    self.format_products_filter(products)
                ],
            },
            'meth': 'StationBoard'
        }

    def parse_station_board_request(
            self: ProfileInterface,
            data: HafasResponse) -> List[Leg]:
        """
        Parses the HaFAS data for the station board request

        :param data: Formatted HaFAS response
        :return: List of journey objects
        """
        legs = []
        try:
            for raw_leg in data.res['jnyL']:
                leg = self.parse_leg(
                    raw_leg,
                    data.res['common'],
                    raw_leg['stopL'][0],
                    raw_leg['stopL'][-1],
                    self.parse_date(raw_leg['date'])
                )
                legs.append(leg)
        except KeyError:
            pass

        return legs



import abc

from pyhafas.types.fptf import Journey
from pyhafas.types.hafas_response import HafasResponse


class JourneyRequestInterface(abc.ABC):
    @abc.abstractmethod
    def format_journey_request(self, journey: Journey) -> dict:
        pass

    @abc.abstractmethod
    def parse_journey_request(self, data: HafasResponse) -> Journey:
        pass
